from flask import request
from flask_restful import Resource
from lyrebird import application


class Checker(Resource):
    def get(self, checker_id=None):
        if not checker_id:
            sorted_checker_groups = application.server['checker'].get_sorted_checker_groups()
            return application.make_ok_response(data=sorted_checker_groups)

        _checker = application.checkers.get(checker_id)
        if not _checker:
            return application.make_fail_response(f'Checker {checker_id} not found')

        content = _checker.read()
        return application.make_ok_response(data=content)

    def put(self, checker_id):
        _checker = application.checkers.get(checker_id)
        if not _checker:
            return application.make_fail_response(f'Checker {checker_id} not found')

        if 'status' not in request.json:
            return application.make_fail_response(f'Checker {checker_id} status not found')

        _checker_status = request.json.get('status')
        if _checker_status:
            application.checkers[checker_id].activate()
        else:
            application.checkers[checker_id].deactivate()
        return application.make_ok_response()

    def post(self, checker_id):
        _checker = application.checkers.get(checker_id)
        if not _checker:
            return application.make_fail_response(f'Checker {checker_id} not found')

        if 'content' not in request.json:
            return application.make_fail_response(f'Checker {checker_id} data not found')

        _checker_data = request.json.get('content')
        origin_checker_detail = _checker.read()
        try:
            _checker.write(_checker_data)
        except Exception as e:
            _checker.write(origin_checker_detail)
            return application.make_fail_response(f'Save checker {checker_id} failure: {str(e)}')

        return application.make_ok_response()



from copy import deepcopy
from lyrebird import application
from .function_executor import FunctionExecutor


class EncoderDecoder(FunctionExecutor):
    def __init__(self):
        self.encoder = application.encoder
        self.decoder = application.decoder

    def encoder_handler(self, flow, output=None):
        matched_funcs = EncoderDecoder.get_matched_handler(self.encoder, flow)
        if output == None:
            EncoderDecoder.func_handler(matched_funcs, flow, handler_type='encoder')
            return

        if not matched_funcs:
            output.update(flow)
            return

        new_flow = deepcopy(flow)
        EncoderDecoder.func_handler(matched_funcs, new_flow, handler_type='encoder')
        output.update(new_flow)

    def decoder_handler(self, flow, output=None):
        matched_funcs = EncoderDecoder.get_matched_handler(self.decoder, flow)
        if output == None:
            EncoderDecoder.func_handler(matched_funcs, flow, handler_type='decoder')
            return

        if not matched_funcs:
            output.update(flow)
            return

        new_flow = deepcopy(flow)
        EncoderDecoder.func_handler(matched_funcs, new_flow, handler_type='decoder')
        output.update(new_flow)




import multiprocessing

from lyrebird import application
from lyrebird.log import get_logger
from .server import serve


logger = get_logger()


class ExtraMockServer():
    def __init__(self) -> None:
        self._server_process = None

    def start(self):
        self._server_process = multiprocessing.Process(
            group=None,
            daemon=True,
            target=serve,
            kwargs={'config': application.config.raw()})
        self._server_process.start()

    def stop(self):
        if self._server_process:
            self._server_process.terminate()
            logger.warning(f'MockServer shutdown')
            self._server_process = None



"""
Script for mitmdump

Redirect request from proxy server to mock server
"""

from urllib.parse import urlparse
from mitmproxy import http
from lyrebird import log
import os
import json
import logging
import re

_logger = log.get_logger()
_logger.setLevel(logging.INFO)

PROXY_PORT = int(os.environ.get('PROXY_PORT'))
PROXY_FILTERS = json.loads(os.environ.get('PROXY_FILTERS'))


def to_mock_server(flow: http.HTTPFlow):
    raw_url = urlparse(flow.request.url)
    raw_host = raw_url.hostname
    if raw_url.port:
        raw_host += f':{raw_url.port}'
    # mock path 为/mock开头加上原始url
    flow.request.path = '/mock/' + flow.request.url
    # mock scheme 统一为http
    flow.request.scheme = 'http'
    # mock server port
    flow.request.port = PROXY_PORT
    # mock server ip
    flow.request.host = '127.0.0.1'
    # device real ip
    address = flow.client_conn.address[0]
    # 获取的address是IPv6（内嵌IPv4地址表示法），需要获取IPv4地址，需要做以下处理
    if address.startswith('::ffff:'):
        address = address.split('::ffff:')[1]
    
    flow.request.headers['Lyrebird-Client-Address'] = address
    flow.request.headers['Mitmproxy-Proxy'] = address
    flow.request.headers['Proxy-Raw-Headers'] = json.dumps({name: flow.request.headers[name] for name in flow.request.headers}, ensure_ascii=False)

    _logger.info('Redirect-> %s' % flow.request.url[:100])


def request(flow: http.HTTPFlow):
    _logger.info(flow.request.url[:100])
    if 'mitm.it' in flow.request.url:
        # Support mitm.it
        return

    if not PROXY_FILTERS:
        to_mock_server(flow)
        return

    for _filter in PROXY_FILTERS:
        if re.search(_filter, flow.request.url):
            to_mock_server(flow)
            break


def responseheaders(flow):
    """
    Enables streaming for all responses.
    This is equivalent to passing `--set stream_large_bodies=1` to mitmproxy.
    """
    if 'mitm.it' in flow.request.url:
        # Support mitm.it
        flow.response.stream = False
        return
    flow.response.stream = True

    command = flow.response.headers.get('Lyrebird-Mitmproxy-Command')
    if command == 'kill':
        flow.kill()



import io
import qrcode
import binascii
from lyrebird import utils

def get_qrcode_img(link):
    formated_link = utils.render(link)

    img = qrcode.make(formated_link, border=1)
    img_buffer = io.BytesIO()
    img.save(img_buffer, 'png')
    img_data_bytes = img_buffer.getvalue()
    img_buffer.close()

    img_data = (b"data:image/png;base64," + binascii.b2a_base64(img_data_bytes)).decode()
    return img_data



from flask_restful import Resource,request
from lyrebird import application
from lyrebird import log
logger = log.get_logger()


class Conf:
    def __init__(self):
        self.bandwidth = -1
        self.bandwidth_templates = [
            {
                "template_name": "UNLIMITED",
                "bandwidth": -1,
                "display": "UNLIMITED"
            },
            {
                "template_name": "2G",
                "bandwidth": 10,
                "display": "2G (10 Kb/s)"
            }, {
                "template_name": "2.5G",
                "bandwidth": 35,
                "display": "2.5G (35 Kb/s)"
            },
            {
                "template_name": "3G",
                "bandwidth": 120,
                "display": "3G (120 Kb/s)"
            }
        ]
        self.indexes_by_name = {
            i["template_name"]: i["bandwidth"] for i in self.bandwidth_templates}
config = Conf()

class Bandwidth(Resource):
    """
    网络带宽
    """
    def get(self):
        return application.make_ok_response(
            bandwidth=config.bandwidth
        )
    def put(self):
        template = request.json["templateName"]
        # check template valid
        if template in list(config.indexes_by_name.keys()):
            # reset bandwidth
            config.bandwidth = config.indexes_by_name[template]
            logger.debug(f'Modified bandwidth: {config.bandwidth}')
            return application.make_ok_response(bandwidth=config.bandwidth)
        else:
            return application.make_fail_response(msg=f'invalid template ! request data is {template}')

class BandwidthTemplates(Resource):
    """
    网络带宽模板列表
    """
    def get(self):
        return application.make_ok_response(
            bandwidth_templates=config.bandwidth_templates
        )




from lyrebird import on_response
from lyrebird.checker import ExtensionCategory

TITLE = '<示例脚本>在Response中添加Header Key'
CATEGORY = ExtensionCategory.MODIFIER

@on_response(rules={
    "request.url": "(?=.*poi/detail)"
})
def add_tag_in_response_headers(flow):
    flow['response']['headers']['Mock-Tag'] = 'Lyrebird'



from aiohttp import web, client
from typing import List, Set, Optional
from urllib import parse as urlparse


class UnknownLyrebirdProxyProtocol(Exception):
    pass


class LyrebirdProxyContext:
    def __init__(self):
        self.request: web.Request = None
        self.netloc = None
        self.origin_url = None
        self.forward_url = None
        self.init = False
        self.protocol_parsers = [
            self.protocol_read_from_path,
            self.protocol_read_from_header,
            self.protocol_read_from_query,
            self.protocol_read_from_query_2
        ]

    def protocol_read_from_path(self, request: web.Request, lb_config):
        # Lyrebird proxy protocol #1
        # Origin target info in path
        # e.g.
        # http://{lyrebird_host}/{origin_full_url}

        if not request.path.startswith('/http://') and not request.path.startswith('/https://'):
            return

        origin_full_url = request.path_qs[1:]
        url = urlparse.urlparse(origin_full_url)

        # origin url for proxy
        self.origin_url = origin_full_url
        # forward url to lyrebird main port 'default 9090'
        port = lb_config.get('mock.port')
        self.forward_url = f'http://127.0.0.1:{port}/mock/{origin_full_url}'

        self.netloc = url.netloc
        self.request = request

        # Set init success
        self.init = True

    def protocol_read_from_header(self, request: web.Request, lb_config):
        # Lyrebird proxy protocol #2
        # Origin target info in headers
        # e.g.
        # http://{lyrebird_host}/{origing_path_and_query}
        # headers:
        # MKScheme: {origin_scheme}
        # MKOriginHost: {origin_host}
        # MKOriginPort: {origin_port}
        header_name_scheme = lb_config.get('mock.proxy_headers', {}).get('scheme')
        header_name_host = lb_config.get('mock.proxy_headers', {}).get('host')
        header_name_port = lb_config.get('mock.proxy_headers', {}).get('port')

        if not request.headers.get(header_name_host) or not request.headers.get(header_name_scheme):
            return

        target_scheme = request.headers.get(header_name_scheme)
        target_host = request.headers.get(header_name_host)
        target_port = request.headers.get(header_name_port)
        if target_port:
            origin_full_url = f'{target_scheme}://{target_host}:{target_port}{request.rel_url}'
            netloc = f'{target_host}:{target_port}'
        else:
            origin_full_url = f'{target_scheme}://{target_host}{request.rel_url}'
            netloc = target_host

        # origin url for proxy
        self.origin_url = origin_full_url
        # forward url to lyrebird main port 'default 9090'
        port = lb_config.get('mock.port')
        self.forward_url = f'http://127.0.0.1:{port}/mock/{origin_full_url}'
        self.netloc = netloc
        self.request = request

        # Set init success
        self.init = True

    def protocol_read_from_query(self, request: web.Request, lb_config):
        # Lyrebird proxy protocol #3
        # Origin target info in query
        # key is "proxy"
        # e.g.
        # http://{lyrebird_host}/{origing_path}?{proxy=encoded-url}

        # default key=proxy
        if not request.query.get('proxy'):
            return

        origin_url = request.query.get('proxy')
        origin_url = urlparse.unquote(origin_url)
        url = urlparse.urlparse(origin_url)

        self.origin_url = origin_url
        # forward url to lyrebird main port 'default 9090'
        port = lb_config.get('mock.port')
        self.forward_url = f'http://127.0.0.1:{port}/mock/?proxy={urlparse.quote(origin_url)}'
        self.netloc = url.netloc
        self.request = request

        # Set init success
        self.init = True

    def protocol_read_from_query_2(self, request: web.Request, lb_config):
        # Lyrebird proxy protocol #4
        # Origin target info in query
        # Set proxyscheme、proxyhost、proxypath
        # e.g.
        # http://{lyrebird_host}?proxyscheme={origin_scheme}&proxyhost={urlencode(origin_host)}&proxypath={urlencode(origin_path)}&{origin_query}
        proxy_scheme = request.query.get('proxyscheme', 'http')
        proxy_host = request.query.get('proxyhost', None)
        proxy_path = request.query.get('proxypath', '/')

        if not proxy_host:
            return

        # remove lyrebrid proxy protocol keys from query string
        origin_query_str = ''
        for query_key, query_value in request.query.items():
            if query_key in ['proxyscheme', 'proxyhost', 'proxypath']:
                continue
            origin_query_str += f'&{query_key}={query_value}'
        if len(origin_query_str) >= 1:
            origin_query_str = '?'+origin_query_str[1:]

        origin_url = f'{proxy_scheme}://{urlparse.unquote(proxy_host)}{urlparse.unquote(proxy_path)}{origin_query_str}'

        self.origin_url = origin_url
        url = urlparse.urlparse(origin_url)

        # forward url to lyrebird main port 'default 9090'
        port = lb_config.get('mock.port')
        self.forward_url = f'http://127.0.0.1:{port}/mock/{origin_url}'
        self.netloc = url.netloc
        self.request = request

        # Set init success
        self.init = True

    @classmethod
    def parse(cls, request: web.Request, lb_config):
        ctx = cls()
        for parser in ctx.protocol_parsers:
            parser(request, lb_config)
            if ctx.init:
                return ctx
        raise UnknownLyrebirdProxyProtocol



from lyrebird.log import get_logger
from lyrebird.mock.dm.jsonpath import jsonpath
from lyrebird.utils import TargetMatch

logger = get_logger()


class Filter:
    
    @staticmethod
    def is_target_match_patterns(pattern_list, target):
        if not pattern_list or not target:
            return False
        for pattern in pattern_list:
            if TargetMatch.is_match(target, pattern):
                return True
        return False
    
    @staticmethod
    def is_jsonpath_match_patterns(pattern_list, json_obj, target_rule):
        target_nodes = jsonpath.search(json_obj, target_rule)
        for node in target_nodes:
            is_match = Filter.is_target_match_patterns(pattern_list, node.node)
            if is_match:
                return True
        return False

    @staticmethod
    def is_flow_match_filter(filter_dict, flow):
        for target_rule, patterns in filter_dict.items():
            if type(patterns) != list:
                patterns = [patterns]
            if not Filter.is_jsonpath_match_patterns(patterns, flow, target_rule):
                return False
        return True
    
    @staticmethod
    def get_items_after_filtration(source_items, filter_obj):
        if not filter_obj:
            return source_items
        advanced_filter = filter_obj.get('advanced', None)
        if advanced_filter:
            must_filter = advanced_filter.get('must', {})
            must_not_filter = advanced_filter.get('must_not', {})
        else:
            must_filter = {}
            must_not_filter = {}
            ignore_host = filter_obj.get('ignore', [])
            if ignore_host:
                must_not_filter = {
                    'request.host': ignore_host
                }
        result_items = []
        for item in source_items:
            if must_not_filter and Filter.is_flow_match_filter(must_not_filter, item):
                continue
            if must_filter and (not Filter.is_flow_match_filter(must_filter, item)):
                continue
            result_items.append(item)
        return result_items
